
#include "mixx.h"
#include <errno.h> 

// ******************************************************************************
// Buffer class (by Michael Pramateftakis)

// Head : First FREE Memoryadress
// Tail : First USED Memoryadress or 0 when buffer empty
// End  : Last USED Memoryadress; undefined when buffer empty

//        tail       end  head                      head       tail     end
//        !             !!                          !          !          !
//        ***************                   ********           ************
// -------------------------------------    -------------------------------------

Buffer::Buffer(unsigned long sz)
{
   startadress=0;
   SetBufferSize(sz);
}

void Buffer::SetBufferSize(unsigned long sz)
{
   sem.Acquire();
	delete[] startadress;
	startadress=new char[sz];
   size=sz;
   endadress=startadress+size-1;
   head=startadress;
   tail=0;
   wrap=FALSE;
   datasize=0;
   sem.Release();
}

void Buffer::Clear(void)
{
   sem.Acquire();
   endadress=startadress+size-1;
   head=startadress;
   tail=0;
   wrap=FALSE;
   datasize=0;
   sem.Release();
}

Buffer::~Buffer()
{
	delete[] startadress;
}

char *Buffer::QueryFree(unsigned long sz)
{
   if (sz>size) return NULL; //Buffer too small
   if (tail==head) return NULL;  //Buffer full
	if (tail==0) return startadress;  //Buffer empty

   if (wrap==TRUE) { head=wrapsavehead; wrap=FALSE; }

   if (head>tail)
   {
   	if ((unsigned long)(endadress-head+1)>=sz) 
         return head;
      else if((unsigned long)(tail-startadress)>=sz)
      {
         wrapsavehead=head;
         wrap=TRUE;
         head=startadress;
      	return head;
      }
   }
   else
   {
   	if ((unsigned long)(tail-head)>=sz) return head;
   }
   return NULL;
}

char *Buffer::QueryValid(unsigned long &sz)
{
   if (tail==0) { sz=0; return NULL; }
	sz=end-tail+1;
   return tail;
}

void Buffer::Brought(unsigned long n)
{
   if(n==0) return;
   sem.Acquire();
	head+=n;
   datasize+=n;
   if(head>tail) end=head-1;
   if(tail==0) tail=startadress;
   if(head>endadress) head=startadress;	//Buffer wrap-around
   wrap=FALSE;
   sem.Release();
   NotifyBrought(n);
}

void Buffer::Taken(unsigned long n)
{
   if(n==0) return;
	sem.Acquire();
   tail+=n;
   datasize-=n;
   if (tail>end) //Buffer now empty
   {
      if(head>startadress && datasize>0)
      {
        	tail=startadress+(tail-end-1);
    		end=head-1;
      }
      else
      {
         tail=0;
         head=startadress;
      }
   }
   sem.Release();
   NotifyTaken(n);
}

int Buffer::GetFillLevel(void)
{
   if (!size) return 0;
   double l=(double)datasize/(double)size;
	return (int)(l*100);
}

void Buffer::NotifyBrought(unsigned long n) { if(n==0) return; }

void Buffer::NotifyTaken(unsigned long n) { if(n==0) return; }

// ******************************************************************************
// Thread & Semaphore classes (by Michael Pramateftakis)

Thread::Thread(void)
{
   HThread=0;
   ThreadId=0;
}

Thread::~Thread()
{
	Terminate();
}

HANDLE Thread::GetThreadHandle()
{
	return HThread;
}

DWORD Thread::GetThreadId()
{
	return ThreadId;
}

int Thread::GetPriority()
{
	int p=GetThreadPriority(HThread);
   if(p==THREAD_PRIORITY_IDLE) return 3;
   if(p==THREAD_PRIORITY_LOWEST) return 7;
   if(p==THREAD_PRIORITY_BELOW_NORMAL) return 11;
   if(p==THREAD_PRIORITY_NORMAL) return 15;
   if(p==THREAD_PRIORITY_ABOVE_NORMAL) return 19;
   if(p==THREAD_PRIORITY_HIGHEST) return 23;
   return 27; //If TIME_CRITICAL.
}

BOOL Thread::SetPriority(int p)
{
   int pr;
   if(p<5) pr=THREAD_PRIORITY_IDLE;
   if((p>=5)&&(p<9)) pr=THREAD_PRIORITY_LOWEST;
   if((p>=9)&&(p<13)) pr=THREAD_PRIORITY_BELOW_NORMAL;
   if((p>=13)&&(p<17)) pr=THREAD_PRIORITY_NORMAL;
   if((p>=17)&&(p<21)) pr=THREAD_PRIORITY_ABOVE_NORMAL;
   if((p>=21)&&(p<25)) pr=THREAD_PRIORITY_HIGHEST;
   if((p>=25)&&(p<30)) pr=THREAD_PRIORITY_TIME_CRITICAL;
	return SetThreadPriority(HThread,pr);
}

BOOL Thread::Suspend()
{
	if (SuspendThread(HThread)!=0xFFFFFFFF) return TRUE;
   return FALSE;
}

BOOL Thread::Resume()
{
	if (ResumeThread(HThread)!=0xFFFFFFFF) return TRUE;
   else return FALSE;
}

BOOL Thread::Terminate(DWORD ec)
{
	if (HThread==0) return TRUE;
   else if(TerminateThread(HThread,ec)==FALSE) return FALSE;
	HThread=0;
   return TRUE;
}

DWORD Thread::TProc(LPDWORD adr)
{
	((Thread *)adr)->Run();
   ExitThread(0);
   ((Thread *)adr)->HThread=0;
   return 0;
}

BOOL Thread::Start()
{
	HThread=CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)TProc,(LPVOID)this,0,&ThreadId);
   if (HThread==NULL) return FALSE;
   return TRUE;
}


Semaphore::Semaphore()
{
	HSem=CreateMutex(NULL,FALSE,NULL);
}

Semaphore::~Semaphore()
{
	if(HSem==NULL) return;
   else CloseHandle(HSem);
}

BOOL Semaphore::Acquire(DWORD timeout)
{
	if(WaitForSingleObject(HSem,timeout)==WAIT_FAILED) return FALSE;
   return TRUE;
}

BOOL Semaphore::Release()
{
	return ReleaseMutex(HSem);
}

HANDLE Semaphore::GetHandle()
{
	return HSem;
}

// ******************************************************************************
// Transport & ConstantRate classes (by Michael Pramateftakis)

int Transport::socket_count=0;

Transport::Transport(int t)
{
   s=INVALID_SOCKET;
   type=t;     // SOCK_DGRAM(UDP;default) oder SOCK_STREAM(TCP)
   lasterr=0;

   socket_count++;
   if (socket_count==1) { WSADATA wsaData; lasterr=WSAStartup(MAKEWORD(1, 1), &wsaData); }
}

Transport::~Transport()
{
   Close();

   socket_count--;
   if (!socket_count) WSACleanup();
}

void Transport::Open(void)
{
   Close();
   memset(&peer, 0, sizeof(struct sockaddr_in));
   memset(&local, 0, sizeof(struct sockaddr_in));
   s=socket(AF_INET, type, 0);
   lasterr=(s==INVALID_SOCKET)?WSAGetLastError():0;
}

void Transport::Close(void)
{
   if (s!=INVALID_SOCKET) closesocket(s);
   s=INVALID_SOCKET;
   lasterr=0;
}

int Transport::Send(char *buf, int siz)
{
 	int r=send(s,buf,siz,0); if (r==SOCKET_ERROR) lasterr=WSAGetLastError();
   return r;
}

int Transport::Receive(char *buf, int sz)
{
	int r=recv(s,buf,sz,0);
   if (r==SOCKET_ERROR) lasterr=WSAGetLastError();
   return r;
}

int Transport::GetLastError(void)
{
	return lasterr;
}

/*
char* Transport::GetLastErrorStr(void)
{
	switch (lasterr)
   {
   case WSANOTINITIALISED: return "A successful WSAStartup must occur before using this function.";
   case WSAENETDOWN: return "The Windows Sockets implementation has detected that the network subsystem has failed.";
   case WSAEADDRINUSE: return "The specified address is already in use. (See the SO_REUSEADDR socket option under setsockopt.)";
   case WSAEFAULT: return "The namelen argument is too small (less than the size of a struct sockaddr).";
   case WSAEINPROGRESS: return "A blocking Windows Sockets call is in progress.";
   case WSAEAFNOSUPPORT: return "The specified address family is not supported by this protocol.";
   case WSAEINVAL: return "The socket is already bound to an address.";
   case WSAENOBUFS: return "Not enough buffers available, too many connections.";
   case WSAENOTSOCK: return "The descriptor is not a socket.";
   }
   return "???";
}
*/

BOOL Transport::Connect(char *host,u_short port)
{
   Open();
	peer.sin_family=AF_INET;
   peer.sin_port=htons(port);
   if (!(*host))
      peer.sin_addr.s_addr=INADDR_BROADCAST;
   else if (inet_addr(host)!=INADDR_NONE) 
      peer.sin_addr.s_addr=inet_addr(host);
   else
   {
   	LPHOSTENT hent=gethostbyname(host); if(!hent) return FALSE;
   	peer.sin_addr.s_addr=*((u_long FAR *)(hent->h_addr));
   }
   if (connect(s,(sockaddr *)&peer,sizeof(struct sockaddr))==SOCKET_ERROR) { lasterr=WSAGetLastError(); return FALSE; }
	return TRUE;
}

BOOL Transport::Connect(u_short port)
{
   Open();
	peer.sin_family=AF_INET;
   peer.sin_port=htons(port);
   peer.sin_addr.s_addr=INADDR_ANY;
   if (connect(s,(sockaddr *)&peer,sizeof(struct sockaddr))==SOCKET_ERROR) { lasterr=WSAGetLastError(); return FALSE; }
	return TRUE;
}

BOOL Transport::Bind(u_short port)
{
   Open();
   local.sin_family=AF_INET;
   local.sin_port=htons(port);
   local.sin_addr.s_addr=INADDR_ANY;
   if (bind(s, (sockaddr *)&local, sizeof(struct sockaddr))==SOCKET_ERROR) { lasterr=WSAGetLastError(); return FALSE; }
   if (type==SOCK_STREAM) if (listen(s, 5)<0) { lasterr=WSAGetLastError(); return FALSE; }
   return TRUE;
}

Transport& Transport::Accept(void)
{
   Transport *t=new Transport(SOCK_STREAM);
   struct sockaddr addr;
   int addrlen=sizeof(struct sockaddr); memset(&addr, 0, addrlen);
   do t->s=accept(s, &addr, &addrlen); while (t->IsInvalid() && errno==EINTR);
   return *t;
}

int Transport::IsInvalid(void)
{
   return (s==INVALID_SOCKET)?TRUE:FALSE;
}

char *Transport::GetPeername(int flg)
{
   int len=sizeof(sockaddr);
   getpeername(s,(sockaddr*)&local,&len);
   if (flg)
	   return (char*)&local.sin_addr;
   else
   {
      hostent *hent=gethostbyaddr((char*)&local.sin_addr, 4, AF_INET);
      char *a=hent->h_name; while (*a && *a!=32) a++; *a=0;
	   return hent->h_name;
   }
}

//----------------------------------------------------------------------------

void CALLBACK CRSend(UINT,UINT,DWORD a,DWORD,DWORD)
{
   static int started;
   if (started) return;
   started=1;
   ConstantRate *b=(ConstantRate *)a;
   unsigned long sz=0;
   char *adr;
   int s;
   unsigned int blocksz=b->GetBlockSize();
	adr=b->GetTransferBuffer()->QueryValid(sz);
   if(sz>0 && adr)
   {
      if(sz>=blocksz)
      {
         s=b->Send(adr,blocksz); 
         if (s==SOCKET_ERROR) { b->OnDisconnect(); started=0; return; }
         b->GetTransferBuffer()->Taken(s);
      }
      else
      {
         s=b->Send(adr,sz); 
         if (s==SOCKET_ERROR) { b->OnDisconnect(); started=0; return; }
         b->GetTransferBuffer()->Taken(s);

         //Second chance if buffer wrapped around;
         adr=b->GetTransferBuffer()->QueryValid(sz);
         if(sz>=(blocksz-s))
         {
            sz=blocksz-s;
         	s=b->Send(adr,sz); 
            if (s==SOCKET_ERROR) { b->OnDisconnect(); started=0; return; }
            b->GetTransferBuffer()->Taken(s);
         }
      }
   }
   started=0;
}

ConstantRate::ConstantRate(UINT rate,UINT blocksize) //rate in Bit/s, blocksize in Bytes
{
   blocksz=blocksize?blocksize:8192;
   r=rate?rate:300000;
   UINT blkps=r/(blocksz*8);     // Blocks per second
   double tim=(1.0/blkps)*1000;
   delay=(UINT)tim;
	timeGetDevCaps(&tc, sizeof(TIMECAPS));
   sending=FALSE;
   thrstarted=FALSE;
   buffer=NULL;
   timestarted=FALSE;
   peername=0;
}

ConstantRate::~ConstantRate()
{
	if (timestarted==TRUE) timeEndPeriod(tc.wPeriodMin);
   if (sending) StopCRSend();
}

void ConstantRate::SetTransferBuffer(class Buffer *buf)
{
	buffer=buf;
}

class Buffer *ConstantRate::GetTransferBuffer()
{
	return buffer;
}

BOOL ConstantRate::ChangeRate(UINT newrate)
{
   if (!newrate || !blocksz) return FALSE;
   UINT blkps=newrate/(blocksz*8);
   if (!blkps) return FALSE;

   r=newrate;
   double tim=(1.0/blkps)*1000;
   delay=(UINT)tim;
	if(sending)
   {
   	if(StopCRSend()==FALSE) return FALSE;
      if(StartCRSend()==FALSE) return FALSE;
   }
   return TRUE;
}

BOOL ConstantRate::ChangeBlockSz(UINT newsz)
{
   if (!newsz) return FALSE;
   blocksz=newsz;
   UINT blkps=r/(newsz*8);
	double tim=(1.0/blkps)*1000;
   delay=(UINT)tim;
	if(sending)
   {
   	if(StopCRSend()==FALSE) return FALSE;
      if(StartCRSend()==FALSE) return FALSE;
   }
   return TRUE;
}

UINT ConstantRate::GetRate()
{
	return r;
}

UINT ConstantRate::GetBlockSize()
{
	return blocksz;
}

BOOL ConstantRate::StartCRSend()
{
   if (sending) return FALSE;
   if (delay==0) return FALSE;
   timeBeginPeriod(tc.wPeriodMin);
   timestarted=TRUE;
	timer=timeSetEvent(delay,0,CRSend,(DWORD)this,TIME_PERIODIC);
   if (timer==NULL) return FALSE;
   sending=TRUE;
   return TRUE;
}

BOOL ConstantRate::StopCRSend()
{
   if (!sending) return TRUE;
	if(timeKillEvent(timer)!=TIMERR_NOERROR) return FALSE;
   else
   {
      sending=FALSE;
      return TRUE;
   }
}

BOOL ConstantRate::StartReceive()
{
	if(thrstarted==FALSE)
   {
   	if(Start()==FALSE) return FALSE;
      thrstarted==TRUE;
   }
   else if(Resume()==FALSE) return FALSE;
   return TRUE;
}

BOOL ConstantRate::StopReceive()
{
	Suspend();
   return Terminate();
}

void ConstantRate::Run()
{
   int onoverflow=0;
   while(1)
      if (type==SOCK_STREAM) 
      {
         client=Accept(); if (client.IsInvalid()) break;
         while(1)
         {
            char *adr=buffer->QueryFree(blocksz);
            if (adr)
            {
	            int r=client.Receive(adr,blocksz); if (r==SOCKET_ERROR || r<=0) break;
               peername=client.GetPeername(1);
               buffer->Brought(r); OnReceive(adr,r);
               onoverflow=0;
            }
            else
               if (!onoverflow) { onoverflow=1; OnOverflow(); }
         }
         client.Close();
      }
      else
      {
         char *adr=buffer->QueryFree(blocksz);
         if (adr)
         {
            struct sockaddr_in from;
            int len=sizeof(sockaddr);
            r=recvfrom(s,adr,blocksz,0,(sockaddr*)&from,&len);
            peername=(char*)&from.sin_addr;
            if (r!=SOCKET_ERROR) { buffer->Brought(r); OnReceive(adr,r); }
            onoverflow=0;
         }
         else
            if (!onoverflow) { onoverflow=1; OnOverflow(); }
      }
}
